# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains the Imgur extractor class"""

import re

from .extractor import Extractor


class MangaSushiEX(Extractor):
    """
    An extractor for MangaSushi.net

    For any URL, get the base page (../manga/NAME) to retrieve the tag data.
    For a series URL, search the page to find chapter URLS
    For a chapter URL, search the page to find 'var chapter_preload_images'
    to get the image URLs.
    """
    def __init__(self):
        super().__init__()
        self.name = "MangaSushi"
        self.url = "https://www.mangasushi.net"
        self._PAGE_PATTERN = r"https?://(?:www\.)?mangasushi\.net.*"
        self._GALLERY_PATTERN = r"https?://(?:www\.)?mangasushi\.net/" + \
                                r"manga/[\w\-]*(?:/[\w\-]*)*/?"
        self._CHAPTER_PATTERN = r"https?://(?:www\.)?mangasushi\.net/" + \
                                r"manga/[\w\-]*/chapter-\d+/p/\d+/?"
        # Map chapter number to URL and soup
        self._all_chapters = {}

    # =========================================================================
    # State checks
    # =========================================================================

    def _is_chapter(self, url):
        """Check if the URL passed is a chapter-type page"""
        return bool(re.match(self._CHAPTER_PATTERN, url))

    # =========================================================================
    # State modifiers
    # =========================================================================

    def reset(self):
        """Reset the all_chapters variable"""
        self._all_chapters = {}

    # =========================================================================
    # Getters
    # =========================================================================

    def get_size(self, url, soup, args):
        if self._is_chapter(url):
            return self._get_chapter_size(soup)
        return self._get_series_size(soup)

    @staticmethod
    def _get_chapter_size(soup):
        """Return the size of a chapter

        The size is found based on the number of <option> tags within a
        specific <select> tag.
        """
        select_tag = soup.find("select", {"id": "single-pager",
                                          "class": "selectpicker"})
        return sum(1 for option in select_tag.findAll("option"))

    def _get_series_size(self, soup):
        """Return the size of the series

        Finds the size by summing the size of all chapters in the series.
        """
        all_chapters = self._all_chapters or self._get_all_chapters(soup)
        return sum(self._get_chapter_size(soup)
                   for _, soup in all_chapters.values())

    def _get_all_chapters(self, soup):
        """Return a dictionary of chapter numbers mapped to URL and soup

        Ex. {"0001": ("http://mangasushi.net/manga/series/chapter-1", soup-1)}
        Also sets the attribute '_all_chapters'
        """
        chapter_tags = soup.findAll("li", {"class": "wp-manga-chapter"})
        for tag in chapter_tags:
            url = tag.a['href']
            chapter = url[-1].zfill(4)
            self._all_chapters[chapter] = (url, self._get_soup(url))

        return self._all_chapters

    def get_gallery_urls(self, url, soup, args):
        if self._is_chapter(url):
            return self._get_chapter_urls(soup)
        return self._get_series_urls(soup)

    @staticmethod
    def _get_chapter_urls(soup):
        """Returns the image URLs and filenames for the chapter

        Finds the image URLs from the 'chapter_preload_images' variable
        from the chapter page
        """
        div_tag = soup.find("div", {"class": "reading-content"})
        js_tag = div_tag.script
        img_vars = re.findall(r"http[^,]*\.jpg", js_tag.text)
        img_urls = (url.replace('\\', '') for url in img_vars)
        return [[url.split('/')[-1], url] for url in img_urls]

    def _get_series_urls(self, soup):
        """Returns the image URLs and filenames for the series

        Combines the image URLs and filenames from each chapter of the series.
        For each chapter, appends the chapter number as a folder to the
        filename.
        """
        urls = []
        all_chapters = self._all_chapters or self._get_all_chapters(soup)
        for chapter_num, (_, ch_soup) in all_chapters.items():
            chapter_imgs = self._get_chapter_urls(ch_soup)
            for filename, img_url in chapter_imgs:
                filename = f"Chapter {chapter_num}/{filename}"
                urls.append([filename, img_url])

        return urls

    def get_tags(self, url, soup, args):
        """The tags are retrieved from the series page"""
        # Get the series URL
        url = '/'.join(url.split('/')[:5])
        series_soup = self._get_soup(url)

        tags = {"Title": self._get_title(series_soup),
                "Languages": "English",
                "Artists": self._get_artists(series_soup),
                "Authors": self._get_authors(series_soup),
                "Genres": self._get_genres(series_soup)}

        # Write chapter tag
        if self._is_chapter(url):
            chapter = re.findall(r"chapter-(\d+)", url)[0]
            tags["Chapters"] = str(chapter).zfill(4)
        else:
            all_chapters = self._all_chapters or self._get_all_chapters(soup)
            chapter_nums = sorted(all_chapters.keys())
            first, last = chapter_nums[0], chapter_nums[-1]
            tags["Chapters"] = f"{first}-{last}"
        return tags

    @staticmethod
    def _get_title(soup):
        return soup.title.string.split('–')[0].strip()

    @staticmethod
    def _parse_tags(soup, div_class):
        """A generic parser for tags, used for getting: artists, authors,
        and genres"""
        div_content = soup.find("div", {"class": div_class})
        tags = div_content.findAll("a")
        return [tag.string for tag in tags]

    @staticmethod
    def _get_artists(soup):
        return MangaSushiEX._parse_tags(soup, "artist-content")

    @staticmethod
    def _get_authors(soup):
        return MangaSushiEX._parse_tags(soup, "author-content")

    @staticmethod
    def _get_genres(soup):
        return MangaSushiEX._parse_tags(soup, "genres-content")



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains the NHentai extractor class"""

import os
import re
from .extractor import Extractor


class NHentaiEX(Extractor):
    """
    An extractor for NHentai.net
    """
    def __init__(self):
        super().__init__()
        self.name = "NHentai"
        self.url = "https://www.nhentai.net"
        self._PAGE_PATTERN = r"https?://(?:t\.)?(?:www\.)?nhentai\.net.*"
        self._GALLERY_PATTERN = r"https?://(?:www\.)?nhentai\.net/g/[0-9]+/?"
        self._IMG_DOMAIN = "https://i.nhentai.net/galleries/"

    # =========================================================================
    # Testing
    # =========================================================================
    def get_tests(self):
        tests = ({"url": "https://nhentai.net/g/252295/",
                  "img_urls": ["https://i.nhentai.net/galleries/1311486/1.jpg",
                               "https://i.nhentai.net/galleries/1311486/2.jpg",
                               "https://i.nhentai.net/galleries/1311486/3.jpg",
                               "https://i.nhentai.net/galleries/1311486/4.jpg",
                               "https://i.nhentai.net/galleries/1311486/5.jpg",
                               "https://i.nhentai.net/galleries/1311486/6.jpg",
                               "https://i.nhentai.net/galleries/1311486/7.jpg",
                               "https://i.nhentai.net/galleries/1311486/8.jpg",
                               "https://i.nhentai.net/galleries/1311486/9.jpg"
                               ],
                  "size": 22,
                  "tags": {"Title": "Ning Hai Nee-chan no Migawari Nikki",
                           "URL": "https://nhentai.net/g/252295/",
                           "Parodies": "Azur Lane",
                           "Characters": ["Teitoku", "Ning Hai"],
                           "Tags": ["Blowjob", "Paizuri", "Swimsuit",
                                    "Bikini", "Lingerie", "Blackmail"],
                           "Artists": "Chiune",
                           "Groups": "Chiukorone",
                           "Languages": "English",
                           "Categories": "Doujinshi"}},

                 {"url": "https://nhentai.net/g/278557/",
                  "img_urls": ["https://i.nhentai.net/galleries/1448240/1.jpg",
                               "https://i.nhentai.net/galleries/1448240/2.jpg",
                               "https://i.nhentai.net/galleries/1448240/3.jpg",
                               "https://i.nhentai.net/galleries/1448240/4.jpg",
                               "https://i.nhentai.net/galleries/1448240/5.jpg",
                               "https://i.nhentai.net/galleries/1448240/6.jpg",
                               "https://i.nhentai.net/galleries/1448240/7.jpg",
                               "https://i.nhentai.net/galleries/1448240/8.jpg",
                               "https://i.nhentai.net/galleries/1448240/9.jpg"
                               ],
                  "size": 18,
                  "tags": {"Title": "姉妹ルフオークの巣窟に行く。",
                           "URL": "https://nhentai.net/g/278557/",
                           "Parodies": "Original",
                           "Artists": "Kekemotsu",
                           "Groups": "Kekemotsu",
                           "Languages": "Japanese",
                           "Categories": "Doujinshi",
                           "Tags": ["Big Breasts", "Group", "Stockings",
                                    "Tentacles", "MMF Threesome", "Orc",
                                    "Monster", "Elf"]}},
                 )

        return tests

    # =========================================================================
    # Getters
    # =========================================================================

    @staticmethod
    def _get_title(soup):
        """Return the title of the gallery

        Title taken from the <h1> tag and may require adjustments.
        """
        raw_title = soup.find("h1").string
        # Remove the brackets and their contents
        square_br = re.sub(r'\[[^\]]*\]', '', raw_title)
        round_br = re.sub(r'\([^\)]*\)', '', square_br)
        curly_br = re.sub(r'\{[^\}]*\}', '', round_br)

        # Remove anything past a '|' char
        bar_strip = re.sub(r'\|.*', '', curly_br)
        title = bar_strip.strip()
        return title

    @staticmethod
    def _get_gallery_id(soup):
        """Return ID of the gallery

        Used to build gallery URLs (ID appended to IMG_DOMAIN)
        """
        url = soup.find(property="og:image")["content"]
        return url.split("/")[4]

    @staticmethod
    def _get_img_type(soup):
        """Return the file format for the images"""
        url = soup.find(property="og:image")["content"]
        return url.split(".")[-1]

    def get_size(self, url, soup, args):
        """Return total # of images in gallery"""
        tags = soup.time.parent.previous_siblings
        for tag in tags:
            if self._is_size_str(tag.string):
                return int(tag.string.split(" ")[0])
        return None

    def get_gallery_urls(self, url, soup, args):
        url_list = []
        size = self.get_size(url, soup, args)
        # Length of gallery size for zero-appending filenames
        size_len = len(str(abs(size)))
        id_ = self._get_gallery_id(soup)
        img_type = self._get_img_type(soup)
        for img_num in range(1, size+1):
            img_url = self._build_img_url(id_, img_num, img_type)
            basename = str(img_num).zfill(size_len)
            filename = f"{basename}.{img_type}"
            url_list.append([filename, img_url])
        return url_list

    def get_tags(self, url, soup, args):
        soup_tags = {"Title": self._get_title(soup), "URL": url}
        tags = soup.find("section", {"id": "tags"})
        tag_divs = tags.find_all("div")
        for div in tag_divs:
            tag_name = re.sub(r"\s+", "", div.contents[0])[:-1]
            tag_content = re.sub(r"^\s+", "", div.get_text().split(":")[1])
            content_elements = div.span.find_all('a')
            tag_content = [element.contents[0] for element in content_elements]
            if tag_content:
                clean_tags = self._clean_tags(tag_name, tag_content)
                soup_tags[tag_name] = clean_tags
        return soup_tags

    # =========================================================================
    # Misc. functions
    # =========================================================================

    def _build_img_url(self, id_, img_num, img_type):
        """Construct the URL of an image"""
        return f"{self._IMG_DOMAIN}{id_}/{img_num}.{img_type}"

    @staticmethod
    def _is_size_str(string):
        """Determine if the string in the HTML tag contains the gallery size"""
        return re.match(r"^[0-9]+ pages$", string)

    @staticmethod
    def _clean_tags(key, contents):
        """Formats the list of contents for usage as values to a key

        Split by a ' (123,123)' format and capitalize all words.
        Also removes any unnecessary values or keys.
        """
        tag_list = [content.strip().title() for content in contents]
        # Prevent "Translated" from being added to the "Languages:" dict
        if key == "Languages":
            if "Translated" in tag_list:
                tag_list.remove("Translated")
            if "Rewrite" in tag_list:
                tag_list.remove("Rewrite")
        return tag_list



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains the MangaDex extractor class"""

import os
import re
import json
from multiprocessing.dummy import Pool as ThreadPool

from komidl.status import yn_prompt
from .extractor import Extractor


class MangaDexEX(Extractor):
    """
    An extractor for MangaDex.org
    """
    def __init__(self):
        super().__init__()
        self.name = "MangaDex"
        self.url = "https://www.mangadex.org"
        self._PAGE_PATTERN = r"https?://(?:www\.)?mangadex\.org.*"
        self._CHAPTER_PATTERN = r"https?://(?:www\.)?mangadex\.org/" + \
                                r"chapter/[0-9]+/[0-9]+/?"
        self._MANGA_PATTERN = r"https?://(?:www\.)?mangadex\.org/" + \
                              r"title/[0-9]+/[a-z-]+/?"
        self._POST_API = "https://mangadex.org/api/?id="
        self._IMG_DOMAIN = "https://mangadex.org/data/"

        self._STATUS_DICT = {1: "Ongoing", 2: "Completed", None: "Unknown"}
        self._GENRE_DICT = {1: "4-Koma", 2: "Action", 3: "Adventure",
                            4: "Award Winning", 5: "Comedy", 6: "Cooking",
                            7: "Doujinshi", 8: "Drama", 9: "Ecchi",
                            10: "Fantasy", 11: "Gender Bender", 12: "Harem",
                            13: "Historical", 14: "Horror", 15: "Josei",
                            16: "Martial Arts", 17: "Mecha", 18: "Medical",
                            19: "Music", 20: "Mystery", 21: "Oneshot",
                            22: "Psychological", 23: "Romance",
                            24: "School Life", 25: "Sci-Fi", 26: "Seinen",
                            27: "Shoujo", 28: "Shoujo Ai", 29: "Shounen",
                            30: "Shounen Ai", 31: "Slice of Life", 32: "Smut",
                            33: "Sports", 34: "Supernatural", 35: "Tragedy",
                            36: "Webtoon", 37: "Yaoi", 38: "Yuri",
                            39: "[no chapters]", 40: "Game", 41: "Isekai"}

        self._manga_data = None
        self._chapter_data = None
        self._chapter_data_list = []

    # ===============================================================================
    # State checks
    # ===============================================================================

    def is_gallery(self, url):
        return self._is_chapter(url) or self._is_manga(url)

    def _is_chapter(self, url):
        return bool(re.match(self._CHAPTER_PATTERN, url))

    def _is_manga(self, url):
        return bool(re.match(self._MANGA_PATTERN, url))

    @staticmethod
    def _is_multilang(m_data):
        """Return number of different languages for all chapters in m_data"""
        lang_list = {m_data["chapter"][ch_id]["lang_code"]
                     for ch_id in m_data["chapter"].keys()}

        return bool(len(lang_list) > 1)

    # ===============================================================================
    # State modifiers
    # ===============================================================================

    def reset(self):
        """Clear all instance variables"""
        self._manga_data = None
        self._chapter_data = None
        self._chapter_data_list = []

    def _retrieve_chapter_data(self, url, chapter_id=None):
        # Add a sleep? Maybe
        """Get the data for a manga chapter and store in a class variable"""
        if not self._chapter_data or not url:
            if not chapter_id:
                chapter_id = self._get_data_id(url)
            data_url = self._POST_API + chapter_id + "&type=chapter"
            response = self._session.get(data_url, verify=False)
            response.raise_for_status()
            self._chapter_data = json.loads(response.content.decode("utf-8"))
        return self._chapter_data

    def _retrieve_manga_data(self, url, manga_id=None):
        """Get the data for a manga and store in a class variable"""
        if not self._manga_data:
            if not manga_id:
                manga_id = self._get_data_id(url)
            data_url = self._POST_API + str(manga_id) + "&type=manga"
            response = self._session.get(data_url, verify=False)
            response.raise_for_status()
            self._manga_data = json.loads(response.content.decode("utf-8"))
        return self._manga_data

    def _retrieve_cdl(self, m_data, args):
        """Retrieve chapter data list"""
        if not self._chapter_data_list:
            pool = ThreadPool(args.thread_size)
            url_list = [(None, ch_id) for ch_id in m_data["chapter"].keys()
                        if m_data["chapter"][ch_id]["lang_code"] in args.lang]
            # If the preferred language is not found
            if not url_list:
                if self._is_multilang(m_data):
                    # If there are multiple languages, then raise an exception
                    raise ValueError(f"Preferred language ({args.lang[0]}) " +
                                     "not found")
                # Otherwise, prompt for download in found language
                print(f"Warning: Preferred language ({args.lang[-1]}) " +
                      "not found")
                dl = yn_prompt("Download gallery in only available language?",
                               skip=args.yes)
                if not dl:
                    raise ValueError(f"Preferred language ({args.lang[0]}) " +
                                     "not found")

                url_list = [(None, ch_id) for ch_id in m_data['chapter']]

            self._chapter_data_list = pool.starmap(self._retrieve_chapter_data,
                                                   url_list)
        return self._chapter_data_list

    # ===============================================================================
    # Getters
    # ===============================================================================

    @staticmethod
    def _get_data_id(url):
        """Get the ID of the for data GET request"""
        return url.split("/")[-3] if url[-1] == "/" else url.split("/")[-2]

    @staticmethod
    def _get_size_data(c_data):
        """From a chapter data, retrieve size"""
        return int(len(c_data["page_array"]))

    @staticmethod
    def _get_title(m_data):
        """Get title of manga from manga_data"""
        return m_data["manga"]["title"]

    @staticmethod
    def _get_artist(m_data):
        return m_data["manga"]["artist"]

    @staticmethod
    def _get_author(m_data):
        return m_data["manga"]["author"]

    def get_size(self, url, soup, args):
        """Get the number of images to download"""
        size = 0
        if self._is_manga(url):
            m_data = self._retrieve_manga_data(url)
            chapter_data_list = self._retrieve_cdl(m_data, args)
            for chapter in chapter_data_list:
                size += self._get_size_data(chapter)
        elif self._is_chapter(url):
            if not self._chapter_data:
                self._retrieve_chapter_data(url)
            size = self._get_size_data(self._chapter_data)
        return int(size)

    def _get_chapter_urls(self, c_data, ch_path=""):
        """Same output as get_gallery_urls, but for a single chapter"""
        url_list = []
        size = self._get_size_data(c_data)
        size_len = len(str(abs(size)))
        for img_num, sub_url in enumerate(c_data["page_array"]):
            img_type = sub_url.split(".")[-1]
            img_name = f"{str(img_num+1).zfill(size_len)}.{img_type}"
            filename = os.path.join(ch_path, img_name)
            if c_data["server"] == "/data/":
                img_url = f"{self._IMG_DOMAIN + c_data['hash']}/{sub_url}"
            else:
                img_url = f"{c_data['server'] + c_data['hash']}/{sub_url}"
            url_list.append([filename, img_url])
        return url_list

    def _get_manga_urls(self, url, args):
        """Same output as get_gallery_urls, but for all chapters of a manga"""
        url_list = []
        m_data = self._retrieve_manga_data(url)
        # Download each chapter's data from the URL
        ch_data_list = self._retrieve_cdl(m_data, args)
        for ch in ch_data_list:
            num = ch["chapter"]
            ch_path = "Chapter " + num.zfill(4)
            url_list += self._get_chapter_urls(ch, ch_path)
        return url_list

    def get_gallery_urls(self, url, soup, args):
        url_list = []
        if self._is_chapter(url):
            if not self._chapter_data:
                self._retrieve_chapter_data(url)
            url_list = self._get_chapter_urls(self._chapter_data)
        elif self._is_manga(url):
            url_list = self._get_manga_urls(url, args)
        return url_list

    def get_tags(self, url, soup, args):
        tags = {"URL": url}
        # Ensure that manga_data is set
        if self._is_chapter(url):
            if not self._chapter_data:
                self._retrieve_chapter_data(url)
            c_data = self._chapter_data
            m_data = self._retrieve_manga_data(url,
                                               manga_id=c_data["manga_id"])
        else:
            # is_manga
            m_data = self._retrieve_manga_data(url)
            c_data = None
            # Retrieve chapter data for _get_langs to work
            self._retrieve_cdl(m_data, args)
        # Start scraping tags
        tags["Title"] = self._get_title(m_data)
        tags["Languages"] = self._get_langs()
        tags["Artists"] = self._get_artist(m_data)
        tags["Authors"] = self._get_author(m_data)
        tags["Chapters"] = self._get_chapters(url, m_data, c_data)
        tags["Category"] = self._get_categories(m_data)
        tags["Status"] = self._get_status(m_data)
        return tags

    def _get_langs(self):
        if self._chapter_data:
            return self._chapter_data["lang_name"]
        return self._chapter_data_list[0]["lang_name"]

    def _get_chapters(self, url, m_data, c_data):
        chapters = ""
        if self._is_chapter(url):
            chapters = c_data['chapter'].zfill(4)
        elif self._is_manga(url):
            last_chapter = sorted(m_data['chapter'].keys(), key=int)[-1]
            end = m_data['chapter'][last_chapter]['chapter'].zfill(4)
            chapters = "0001-" + end
        return chapters

    def _get_categories(self, m_data):
        return [self._GENRE_DICT[genre] for genre in m_data["manga"]["genres"]]

    def _get_status(self, m_data):
        return self._STATUS_DICT[m_data["manga"]["status"]]



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains the Hentai2Read extractor class"""

import os
import re

from .extractor import Extractor


class Hentai2ReadEX(Extractor):
    """
    An extractor for Hentai2Read.com

    The image URLs come from a JavaScript variable embedded in the first page
    of the gallery (gData). Another quirk is that galleries on this site are
    spread throughout multiple pages.
    """
    def __init__(self):
        super().__init__()
        self.name = "Hentai2Read"
        self.url = "https://www.hentai2read.com"
        self._PAGE_PATTERN = r"https?://(?:www\.)?hentai2read\.com.*"
        self._GALLERY_PATTERN = r"https?://(?:www\.)?hentai2read\.com/[^/]*"
        self._IMG_DOMAIN = "https://static.hentaicdn.com/hentai"

    # =========================================================================
    # Getters
    # =========================================================================

    @staticmethod
    def _get_title(soup):
        title = soup.title.string
        return title.split("(")[0].strip()

    @staticmethod
    def _get_gallery_size(datavar):
        """Returns the total number of images in gallery"""
        for line in datavar:
            if "images" in line:
                return len(line.split(","))-1
        return None

    def _get_datavar(self, url):
        """Returns a list of all items in the gData variable

        gData is a variable in the site's JavaScript and is retrieved from the
        gallery page. It is useful for finding the gallery size and getting
        the image URLs.
        """
        # Download the first gallery page
        url = os.path.join(url, "1")
        soup = self._get_soup(url)
        # Get all instances of <script>
        scripts = soup.find_all("script")
        # Look for the tags that contain content
        #   (assume var gData is the first inst.)
        data = [tag for tag in scripts if len(tag.contents) > 0][0]
        # Strip the content, then split and strip each line to make a list
        return [line.strip() for line in data.contents[0].strip().splitlines()]

    def get_size(self, url, soup, args):
        datavar = self._get_datavar(url)
        return self._get_gallery_size(datavar)

    def get_gallery_urls(self, url, soup, args):
        datavar = self._get_datavar(url)

        url_list = []
        size = self._get_gallery_size(datavar)
        size_len = len(str(abs(size)))
        for line in datavar:
            if "images" in line:
                data_array = line.split(" ")[2]
                url_array = data_array.split(",")[:-1]

        for item in url_array:
            url = self._build_url(item)
            filename = self._build_filename(url, size_len)
            url_list.append([filename, url])
        return url_list

    def get_tags(self, url, soup, args):
        soup_tags = {"Title": self._get_title(soup), "URL": url}
        tags = soup.find_all("li", {"class": "text-primary"})
        for tag in tags:
            if tag.b and tag.find_all("a"):
                key = tag.b.string
                key += "s" if key[-1] != "s" else key
                items = self._clean_tags(tag)
                soup_tags[key] = items
        return soup_tags

    # =========================================================================
    # Misc. functions
    # =========================================================================

    def _build_url(self, item):
        """Return a URL for an item

        Removes any quotes, brackets, and backslashes.
        """
        image = re.sub(r'[\\"\[\]]', "", item)
        return self._IMG_DOMAIN + image

    @staticmethod
    def _build_filename(url, len_size):
        """Return a filename based on the image's URL and gallery size"""
        file_ = url.split("/")[-1]
        ext = file_.split(".")[-1]
        name = file_.split(".")[0][-len_size:]
        return f"{name}.{ext}"

    @staticmethod
    def _clean_tags(tags):
        """
        Build list from tags for the dictionary and format all entries
        """
        items = [item.string for item in tags.find_all("a")]
        # Remove empty "-" placeholder used by site
        items = ["" if item == "-" else item for item in items]
        # Remove newline char
        items = [re.sub(r"\n", "", item) for item in items]
        # Remove bracket content
        items = [item.split("(")[0].strip()
                 if "(" in item else item for item in items]
        return items



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains the Tsumino extractor class"""

import re
import json

from urllib.parse import quote
import requests

from komidl.exceptions import ExtractorFailed
from .extractor import Extractor


class TsuminoEX(Extractor):
    """
    An extractor for Tsumino.com

    Unfortunately, this website tracks requests based on galleries and will
    require the completion of captcha. As a result, the user will have to
    manually visit the website every ~4 galleries.

    All galleries hosted appear to be in English and use JPGs.
    """
    def __init__(self):
        super().__init__()
        self.name = "Tsumino"
        self.url = "https://www.tsumino.com"
        self._PAGE_PATTERN = r"https?://(?:www\.)?tsumino\.com.*"
        self._GALLERY_PATTERN = (r"https?://(?:www\.)?tsumino\.com/"
                                 r"[Bb]ook/[Ii]nfo/[0-9]+/[^/]+/?")
        self._IMG_DOMAIN = "https://www.tsumino.com/Image/Object?name="
        requests.packages.urllib3.disable_warnings()

    # =========================================================================
    # Getters
    # =========================================================================

    @staticmethod
    def _get_title(soup):
        raw_title = soup.title.string
        # Remove the " | Tsumino" bit
        index = [pos for pos, char in enumerate(raw_title) if char == "|"][-1]
        full_title = raw_title[0:index].strip()
        # Note: Format goes: Romaji | English | Japanese
        # Sometimes the English doesn't exist, so Romaji becomes English
        romaji = full_title.split("|")[0].strip()
        # Note the second value is Japanese (eng, jpn = romaji.split('/'))
        english, _ = romaji.split("/")
        return english.strip()

    @staticmethod
    def _get_gallery_id(url):
        if url[-1] == "/":
            return url.split("/")[-3]
        return url.split("/")[-2]

    def get_size(self, url, soup, args):
        size_element = soup.find("div", {"class": "book-data", "id": "Pages"})
        size = size_element.string.strip()
        return int(size)

    def _get_img_urls(self, id_):
        """A POST operation to retrieve/construct image URLs"""
        referer = f"http://www.tsumino.com/Read/View/{id_}"
        self._session.headers.update({"Referer": referer})
        response = self._session.post("http://www.tsumino.com/Read/Load",
                                      data={'q': id_})
        if response.status_code == 404:
            raise ExtractorFailed("Captcha error - visit the site and "
                                  "complete the captcha or try again later")
        data = json.loads(response.content.decode("utf-8"))
        return [self._format_img_url(url) for url in data["reader_page_urls"]]

    def get_gallery_urls(self, url, soup, args):
        url_list = []
        size = self.get_size(url, soup, args)
        # Length of gallery size for zero-appending filenames
        size_len = len(str(size))
        id_ = self._get_gallery_id(url)
        img_urls = self._get_img_urls(id_)
        for img_num in range(1, size+1):
            base_name = str(img_num).zfill(size_len)
            filename = f"{base_name}.jpg"
            url_list.append([filename, img_urls[img_num-1]])
        return url_list

    def get_tags(self, url, soup, args):
        soup_tags = {"Title": self._get_title(soup), "URL": url}
        soup_tags["Tags"] = self._get_content_tags(soup)
        soup_tags["Category"] = self._get_category_tags(soup)
        soup_tags["Languages"] = "English"
        soup_tags["Artists"] = self._get_artist_tags(soup)
        soup_tags["Group"] = self._get_group_tags(soup)
        soup_tags["Parody"] = self._get_parody_tags(soup)
        soup_tags["Characters"] = self._get_character_tags(soup)
        return soup_tags

    @staticmethod
    def _get_content_tags(soup):
        content = soup.find("meta", {"name": "description"})["content"]
        tags = content.split(":")[-1].split(",")
        return [item.strip() for item in tags]

    @staticmethod
    def _get_category_tags(soup):
        return soup.find("a", {"data-type": "Category"})["data-define"]

    @staticmethod
    def _get_artist_tags(soup):
        artist_list = soup.find_all("a", {"data-type": "Artist"})
        return [artist["data-define"] for artist in artist_list]

    @staticmethod
    def _get_group_tags(soup):
        if soup.find("a", {"data-type": "Group"}):
            group_list = soup.find_all("a", {"data-type": "Group"})
            return [group["data-define"] for group in group_list]
        return []

    @staticmethod
    def _get_parody_tags(soup):
        if soup.find("a", {"data-type": "Parody"}):
            parody_tag = soup.find("a", {"data-type": "Parody"})
            return parody_tag["data-define"]
        return []

    @staticmethod
    def _get_character_tags(soup):
        if soup.find("a", {"data-type": "Character"}):
            char_list = soup.find_all("a", {"data-type": "Character"})
            return [char["data-define"] for char in char_list]
        return []

    # ===============================================================================
    # Misc. functions
    # ===============================================================================

    def _format_img_url(self, url):
        clean_url = quote(url)
        return f"{self._IMG_DOMAIN}{clean_url}"

    @staticmethod
    def _clean_tags(key, string):
        """
        Split by the ' (123,123)' type thing
        Capitalize every word
        """
        tags = re.split(r" \([0-9,]+\)", string)[:-1]
        tag_list = [tag.title() for tag in tags]
        # Prevent "Translated" from being added to the "Languages:" dict
        if key == "Languages":
            if "Translated" in tag_list:
                tag_list.remove("Translated")
            if "Rewrite" in tag_list:
                tag_list.remove("Rewrite")
        return tag_list



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains constants used by KomiDL"""

VERSION = "0.5.5"

# Length of statusbar
STATUSBAR_LEN = 65

USER_AGENT = "komidl/0.1"

# Default formats provided by shutil
ARCHIVE_FORMATS = {"zip", "tar", "gztar", "bztar"}

# Strictly static image formats, used by the PDF export
IMAGE_FORMATS = {"jpg", "png", "jpeg", "tiff", "bmp", "webp", "svg"}

# Common image/video formats that can be downloaded by KomiDL
#   - Ordered by expected frequency/regularity
COMMON_FORMATS = {"jpg", "png", "gif", "gifv", "jpeg", "tiff", "bmp", "webp",
                  "svg"}

# ISO 639-1 Codes
LANG_TO_ISO = {"Japanese": "JA", "English": "EN", "Chinese": "ZH",
               "French": "FR", "Spanish": "ES", "Korean": "KR", "German": "DE",
               "Russian": "RU", "Italian": "IT", "Portuguese": "PT"}

# ISO 639-1 Codes to common abbreviations
ISO_LANG_SET = {"ja": ("ja", "jp", "jap",), "en": ("en", "eng", "us", "gb",),
                "zh": ("zh", "cn",), "es": ("es", "mx", "sp",),
                "de": ("de", "ger",), "ru": ("ru", "rus",),
                "pt": ("pt", "br",)}

KOMIDL_LOGO = ("   +++++++++++++++++++++++++++++   \n"
               " +++:-------------------------:+++ \n"
               "+++`.//+++++++++++++++++++++//.`+++\n"
               "++/`+/          `+`          /+`/++\n"
               "++/`++::::::::- `+:::::::::::++`/++\n"
               "++/`++++++++++: `+.          /+`/++\n"
               "++/`++////////- `+:----------++`/++\n"
               "++/`+/          `+.          /+`/++\n"
               "++/`++:::::::::::+:::////////++`/++\n"
               "++/`+/          `+` :++++++++++`/++\n"
               "++/`++--//////- `+` :++++++++++`/++\n"
               "++/`+/  /+++++: `+` :++++++++++`/++\n"
               "++/`+/  ::::::- `+` -::::::::++`/++\n"
               "++/`+/          `+`          /+`/++\n"
               "+++`./+++++++++++++++++++++++/.`+++\n"
               " +++:-------------------------:+++ \n"
               "   +++++++++++++++++++++++++++++   ")



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains the Scraper class"""

import os
import sys
import imghdr
import shutil
from multiprocessing.dummy import Lock
from multiprocessing.dummy import Pool as ThreadPool

import requests
from bs4 import BeautifulSoup

import komidl.status as status
import komidl.constants as constants
from komidl.exceptions import ExtractorFailed, InvalidURL


class Scraper:
    """Download all images from a URL with an extractor.

    The Scraper class given a URL and extractor will work to create directories
    and download the images, as well as write tags to files. File downloading
    is multi-processed and thus requires a mutex lock in order to update
    _downloaded (tracks current download progress for the URL) and to print
    the current status to the screen.
    Scraper also has a session object that is shared among methods, and
    must be used for all web requests.
    """
    def __init__(self, extractor=None):
        self.extractor = extractor
        self._lock = Lock()
        self._session = requests.Session()
        self._session.headers = {"User-Agent": constants.USER_AGENT,
                                 "Accept-encoding": "gzip, deflate"}
        requests.packages.urllib3.disable_warnings()
        self._gallery_size = 0
        self._downloaded = 0

    @staticmethod
    def _soup_request(session: requests.Session, url: str):
        """Returns a BS4 soup from the URL's response

        The request is done using the session.
        """
        request = session.get(url, verify=False)
        request.raise_for_status()
        content = request.content
        return BeautifulSoup(content, "html.parser")

    def reset(self) -> None:
        """Resets the state of the Scraper after use"""
        # Reset gallery size
        self._gallery_size = 0
        # Rebuild lock if dereferenced after InvalidURL exception
        if not self._lock:
            self._lock = Lock()

    @staticmethod
    def _get_extension(url: str):
        """Returns the path and extension of a URL.

        Acts exactly like os.path.splitext(), but without the '.' char in
        the extension.
        """
        path, ext = os.path.splitext(url)
        return path, ext[1:]

    @staticmethod
    def _write_image(img: str, dest: str) -> None:
        """Saves the response of an image request to the dest path."""
        with open(dest, "wb") as file_:
            for chunk in img.iter_content(chunk_size=1024):
                file_.write(chunk)

    def _retry_download_image(self, filename: str, url: str) -> None:
        """Retries downloading an image by using alternate extensions/formats

        Alternate image formats are defined in constants.COMMON_FORMATS.

        Raises an ExtractorFailed exception if all attempts using alternate
        image formats have been exhausted.
        """
        current_ext = self._get_extension(url)
        other_exts = (ext for ext in constants.COMMON_FORMATS
                      if ext != current_ext)
        # Exhaustively try other image formats
        for ext in other_exts:
            new_url = self._change_extension(url, ext)
            response = self._session.get(new_url, stream=True, verify=False)
            if response.status_code == 200:
                self._write_image(response, filename)
                break
        else:
            raise ExtractorFailed(f"Extractor returned an invalid URL: {url}")

    def _download_image(self, filename: str, url: str) -> None:
        """Downloads an image from the URL and saves it to filename.

        Failure to download an image (HTTP 404) may be caused by a wrong
        extension from the extractor. On failure, various other extensions are
        exhaustively tried. If the image can't be downloaded, then the gallery
        should be failed as well.

        Downloading an image updates the status bar, and thus the mutex lock
        is acquired before incrementing and drawing.

        In case the 'URL' or 'filename' argument was incorrect, after a
        successful download the magic bytes of the image is checked and the
        image's file extension may be modified.

        Returns the full path of the downloaded image
        """
        img = self._session.get(url, stream=True, verify=False)
        if img.status_code == 200:
            self._write_image(img, filename)
        elif img.status_code == 404:
            self._retry_download_image(filename, url)
        else:
            raise InvalidURL(f"Server error encountered at: {url}")

        self._lock.acquire()
        self._downloaded += 1
        status.progress_draw(constants.STATUSBAR_LEN, self._downloaded,
                             self._gallery_size)
        self._lock.release()

        return self._fix_extension(filename)

    @staticmethod
    def _fix_extension(image: str) -> str:
        """Renames the image's file extension based on the magic bytes found.

        If magic bytes could not be found, the image's path is not modified.

        Returns image's path with true file extension
        """
        cur_ext = Scraper._get_extension(image)
        actual_ext = imghdr.what(image)
        if actual_ext is not None:
            if cur_ext != actual_ext:
                new_image = Scraper._change_extension(image, actual_ext)
                os.rename(image, new_image)
                return new_image
        return image

    @staticmethod
    def _change_extension(url: str, ext: str) -> str:
        """Replaces the URL's file extension."""
        base_url, _ = os.path.splitext(url)
        return f"{base_url}.{ext}"

    @staticmethod
    def _create_dir(title: str, root_dir: str, overwrite=False) -> str:
        """Creates a directory to hold all downloaded files.

        The directory is created in the root directory, which is defined by
        the --directory argument (by default it is where the user ran the
        script).
        If the root directory does not exists, an exception is raised.
        If 'overwrite' is set, then any existing folder is automatically
        overwritten. Otherwise, a prompt will appear to give the user
        the option to overwrite or create the directory with a new name.

        Returns the full path of the newly created directory.
        """
        root_dir = os.path.abspath(root_dir)
        dest = os.path.join(root_dir, title)

        try:
            os.mkdir(dest)
        except FileExistsError:
            if not overwrite:
                prompt_msg = f"{dest} already exists. Overwrite?"
                overwrite = status.yn_prompt(prompt_msg)
            if overwrite:
                shutil.rmtree(dest)
            else:
                duplicates = sum(1 for dir_ in os.listdir(root_dir)
                                 if title in dir_)
                dest = f"{dest} ({duplicates})"
            os.mkdir(dest)

        return dest

    @staticmethod
    def _append_path(path: str, gallery_urls):
        """Appends the path to all image paths in gallery_urls."""
        for img, url in gallery_urls:
            yield os.path.join(path, img), url

    @staticmethod
    def _create_subdirs(path, gallery_urls) -> None:
        """Create all sub-directories from paths in gallery_urls."""
        img_urls = Scraper._append_path(path, gallery_urls)
        img_paths = (os.path.split(img) for img, _ in img_urls)
        for sub_dir, _ in img_paths:
            os.makedirs(sub_dir, exist_ok=True)

    def scrape(self, url: str, args) -> str:
        """Scrapes an image gallery at the URL.

        Using the extractor set within the object, if the URL given is a
        gallery then it is used to scrape.

        Raises an InvalidURL exception if the given URL or scraped image URL
        could not be requested.

        Raises an ExtractorFailed exception if an error was encountered within
        the extractor.

        Raises a ValueError exception if any argument values are incorrect.

        Returns the full path of the directory containing all scraped images.
        """
        self._downloaded = 0

        soup = self._soup_request(self._session, url)

        if not self.extractor.is_gallery(url):
            raise InvalidURL(f"'{url}' is not a valid image gallery for the "
                             f"'{self.extractor.name}' extractor")
        try:
            directory = self.scrape_gallery(soup, url, args)
        finally:
            self.extractor.reset()

        return directory

    def _fail(self, path=None) -> None:
        """Print a failure message and delete all loose files.

        The mutex lock for Scraper is destroyed to prevent further writing by
        the downloading sub-processes.
        """
        self._lock.acquire()
        status.progress_fail(constants.STATUSBAR_LEN, self._downloaded,
                             self._gallery_size)
        self._lock = None
        if path is not None:
            shutil.rmtree(path)

    def _download_images(self, path, urls, thread_size: int) -> None:
        """Download images from the urls to the path"""
        img_urls = self._append_path(path, urls)
        pool = ThreadPool(thread_size)
        try:
            pool.starmap(self._download_image, img_urls)
        finally:
            pool.close()
            pool.join()

    def scrape_gallery(self, soup, url, args) -> str:
        """Scrape a URL and download all images.

        Tags are written to a file if the --tags option is selected.

        In the process, the directory to hold all scraped info is created.
        """
        # Gather info for progress bar img downloading, create dest directory
        tags = self.extractor.get_tags(url, soup, args)
        title = self._build_title(tags)
        path = self._create_dir(title, args.directory, overwrite=args.yes)
        _, dirname = os.path.split(path)
        self._gallery_size = self.extractor.get_size(url, soup, args)

        # Start scraping process
        status.progress_start(dirname)
        try:
            gallery_urls = self.extractor.get_gallery_urls(url, soup, args)
            # Create the sub-directories if needed
            self._create_subdirs(path, gallery_urls)
            # Start downloading images
            self._download_images(path, gallery_urls, args.thread_size)
        except (ExtractorFailed, InvalidURL, KeyboardInterrupt) as e:
            self._fail(path=path)
            raise e
        # Write tags
        if args.tags:
            self._write_tags(tags, path)
        status.progress_finish(constants.STATUSBAR_LEN)

        return path

    @staticmethod
    def _build_title(tags: dict) -> str:
        """Build a name for the directory containing downloaded images"""
        def langs_tostr(langs: str) -> str:
            """Return language as an ISO 639-1 abbreviation"""
            # Get the first language if the value is a list
            if isinstance(langs, list):
                language, *_ = langs
            else:
                language = langs
            return constants.LANG_TO_ISO[language.title()]

        def build_credit_tag(tags: dict) -> str:
            """Credit the author/artist/group of the gallery for the title

            Priority of: Authors -> Artists -> Groups -> UNKNOWN
            """
            credit = tags.get('Authors',
                              tags.get('Artists',
                                       tags.get('Groups',
                                                'Unknown')))
            if isinstance(credit, list):
                return "x".join(name.upper() for name in credit)
            return credit.upper()

        title = tags.get('Title', 'UNTITLED')
        language = langs_tostr(tags['Languages'])
        credit = build_credit_tag(tags)
        chapters = tags.get('Chapters')
        chapter_tag = f"[{chapters}] " if chapters else " "
        return f"[{credit}][{language}]{chapter_tag}{title}"

    @staticmethod
    def _write_tags(tags, path):
        """Writes tags to a text file path"""
        info_str = Scraper._tags_tostr(tags)
        info_file = os.path.join(path, "info.txt")
        with open(info_file, "w") as file_:
            file_.write(info_str)

    @staticmethod
    def _tags_tostr(tags):
        """Returns tags as a string of format: 'KEY:ITEM,ITEM,ITEM'"""
        valid_keys = [key for key in tags if tags[key]]
        item_strs = (','.join(tags[key]) if isinstance(tags[key], list)
                     else tags[key] for key in valid_keys)
        tag_strs = (f"{key}:{item_str}\n"
                    for key, item_str in zip(valid_keys, item_strs))
        return "".join(tag_strs)



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains functions for drawing status messages"""

import sys


def _get_bars(bar_size: int, current: int, total: int) -> int:
    """Calculate the number of bars to display in the progress bar.

    Parameters
    ----------
    bar_size:   Total # of bars to be displayed
    current:    Current units
    total:      Total units required

    Returns
    -------
    The number of bars to display in the progress bar
    """
    return int((current*bar_size)/float(total))


def _get_percent(current: int, total: int) -> int:
    """Calculate the percentage of 'current' from 'total'.

    Parameters
    ----------
    current:    Current units
    total:      Total units required

    Returns
    -------
    The percentage of pages downloaded
    """
    return int(current*100/float(total))


def _progress_bar(string: str) -> None:
    """Print the progress bar message in-place"""
    sys.stdout.write("\r")
    sys.stdout.write(string)
    sys.stdout.flush()


def progress_start(title: str) -> None:
    """Print the starting download message.

    Parameters
    ----------
    title:      The title of the gallery that will be downloaded
    """
    sys.stdout.write(f"Downloading: {title}\n")


def progress_draw(bar_size: int, current: int, total: int) -> None:
    """Draw a progress bar.

    Parameters
    ----------
    bar_size:   Total # of bars to be displayed
    current:    Current units
    total:      Total units required
    """
    bars = _get_bars(bar_size, current, total)
    bar_str = '='*bars
    percent = _get_percent(current, total)
    _progress_bar(f"Progress:    [{bar_str:<{bar_size}}]{percent:>4}%")


def progress_fail(bar_size: int, current: int, total: int) -> None:
    """Draw a progress bar failing at a certain point.

    Parameters
    ----------
    bar_size:   Total # of bars to be displayed
    current:    Current units
    total:      Total units required
    """
    bars = _get_bars(bar_size, current, total)
    bar_str = '='*bars
    percent = _get_percent(current, total)
    _progress_bar(f"FAILED:      [{bar_str:<{bar_size}}]{percent:>4}%")


def progress_stop(bar_size: int, current: int, total: int) -> None:
    """Draw a progress bar stopped at a certain point.

    Parameters
    ----------
    bar_size:   Total # of bars to be displayed
    current:    Current units
    total:      Total units required
    """
    bars = _get_bars(bar_size, current, total)
    bar_str = '='*bars
    percent = _get_percent(current, total)
    _progress_bar(f"STOPPED:      [{bar_str:<{bar_size}}]{percent:>4}%")


def progress_finish(bar_size: int) -> None:
    """Draw a complete progress bar appended by a newline."""
    bar_str = '='*bar_size
    _progress_bar(f"Complete:    [{bar_str:<{bar_size}}]{100:>4}%")
    sys.stdout.write("\n")


def yn_prompt(prompt_msg: str, skip=False) -> bool:
    """Print a message with yes/no prompt and return user input.

    Parameters
    ----------
    prompt_msg : str
        A message/question to accompany the yes/no prompt

    Returns
    -------
    bool
        True if user entered yes, false if otherwise
    """
    if skip:
        return skip

    prompt_msg += " [y/N] "
    response = input(prompt_msg)
    return response.lower() == "y"



"""Tests the Scraper class in KomiDL"""

import os
import sys
import unittest
from unittest import mock

from multiprocessing.dummy import Lock
sys.path.append(os.path.abspath('..'))

from komidl.scraper import Scraper
from komidl.exceptions import ExtractorFailed, InvalidURL


class ScraperTest(unittest.TestCase):
    """Tests the Scraper class in KomiDL"""
    def setUp(self):
        """Create the Scraper object for usage"""
        self.scraper = Scraper(3)

    ####################################################################### 
    # Miscellaneous
    #######################################################################

    def test_change_extension(self):
        """Test the function that changes file ext for a URL"""
        url = "http://website.com/test/image.jpg"
        ext = "png"
        expected = "http://website.com/test/image.png"
        actual = self.scraper._change_extension(url, ext)
        self.assertEqual(actual, expected)

    def test_reset(self):
        self.scraper._gallery_size = 40
        self.scraper._lock = None
        self.scraper.reset()
        self.assertEqual(self.scraper._gallery_size, 0)
        self.assertTrue(self.scraper._lock is not None)

    def test_get_extension(self):
        test_cases = (
                      # Test 'best-case' scenario
                      ('img.jpg', ('img', 'jpg')),
                      # Test image with '.' in middle
                      ('thing.img.gif', ('thing.img', 'gif')),
                     )
        for img, (expected_path, expected_ext) in test_cases:
            (actual_path, actual_ext) = self.scraper._get_extension(img)
            self.assertEqual(actual_path, expected_path)
            self.assertEqual(actual_ext, expected_ext)

    def test_build_title(self):
        test_cases = (
                      (# Test a 'best-case' scenario
                       '[DNSHENG][EN] Some Title',
                       {
                        'Title': 'Some Title',
                        'Languages': 'English',
                        'Authors': 'DNSheng'
                       }
                      ),
                      (# Test no title and multiple languages
                       '[PICASSO][IT] UNTITLED',
                       {
                        'Languages': ['Italian', 'English'],
                        'Artists': 'Picasso'
                       }
                      ),
                      (# Test no credit and chapters
                       '[UNKNOWN][JA][0001-0004] Cool stories',
                       {
                        'Title': 'Cool stories',
                        'Languages': ['Japanese', 'English'],
                        'Chapters': '0001-0004',
                       }
                      ),
                      (# Test group credit
                       '[MINUTEMEN][JA] A settlement needs your help',
                       {
                        'Title': 'A settlement needs your help',
                        'Languages': ['Japanese', 'English'],
                        'Groups': 'Minutemen'
                       }
                      ),
                      (# Test multiple credits
                       '[BOB ROSSxPICASSO][JA] Picture gallery',
                       {
                        'Title': 'Picture gallery',
                        'Languages': ['Japanese', 'English'],
                        'Artists': ['Bob Ross', 'Picasso']
                       }
                      ),

                      (# Test credit priority
                       '[SHAKESPEARE][EN] Romeo & Juliet',
                       {
                        'Title': 'Romeo & Juliet',
                        'Languages': 'English',
                        'Authors': 'Shakespeare',
                        'Artists': 'Bob Ross',
                        'Groups': 'Dead Poets Society',
                       }
                      ),
                     )
        for expected, tag_dict in test_cases:
            actual = self.scraper._build_title(tag_dict)
            self.assertEqual(actual, expected)

    ####################################################################### 
    # _download_image()
    #######################################################################
    @mock.patch('komidl.status.progress_draw')
    def test_download_image(self, mock_draw):
        test_cases = (
                      (# Test a valid image URL
                       '001.png',
                       'http://gallery.com/g/abc/001.png',
                       '001.png'
                      ),
                      (# Test a URL that needs to be re-tried
                       '001.jpg',
                       'http://gallery.com/g/abc/001.jpg',
                       '001.png'
                      ),
                     )

        # Set up the mock functions
        def mock_retry_request(*args, **kwargs):
            # Mock the session.get, only an 'png' img  without 'exception'
            # in the URL will return 200
            class MockResponse:
                def __init__(self, status_code):
                    self.status_code = status_code
            ext = args[0].split('.')[-1]
            if ext == 'png':
                return MockResponse(200)
            return MockResponse(404)

        def mock_fix_extension(filename):
            path = '.'.join(filename.split('.')[:-1])
            return f"{path}.png"

        mock_do_nothing = mock.Mock(return_value=None)

        self.scraper._session.get = mock_retry_request
        self.scraper._write_image = mock_do_nothing
        self.scraper._retry_download_image = mock_do_nothing
        self.scraper._fix_extension = mock_fix_extension

        for path, url, expected in test_cases:
            actual = self.scraper._download_image(path, url)
            self.assertEqual(actual, expected)

    def test_download_image_exception(self):
        test_cases = (
                      (# Test an image URL returning 429
                       '001.png',
                       'http://gallery.com/g/abc/001.png',
                       ExtractorFailed
                      ),
                      (# Test an image URL returning 429
                       '001.png',
                       'http://exception.com/g/abc/001.png',
                       InvalidURL
                      ),
                     )

        # Set up the mock functions
        def mock_retry_request(*args, **kwargs):
            # Returns a mock 429 response for all URLs
            # Returns a mock 404 response otherwise
            class MockResponse:
                def __init__(self, status_code):
                    self.status_code = status_code
            if 'exception' in args[0]:
                return MockResponse(429)
            return MockResponse(404)

        self.scraper._session.get = mock_retry_request

        for path, url, expected in test_cases:
            self.assertRaises(expected,
                              self.scraper._download_image,
                              path,
                              url)

    ####################################################################### 
    # _retry_download_image()
    #######################################################################
    def test_retry_download_image(self):
        test_cases = (
                      (# Test retrying until an 'svg' image is found
                       '001.jpg',
                       'http://gallery.com/g/abc/001.jpg',
                       None
                      ),
                     )

        # Set up the mock functions
        def mock_retry_request(*args, **kwargs):
            # Mock the session.get, only an 'svg' img will return 200
            class MockResponse:
                def __init__(self, status_code):
                    self.status_code = status_code
            ext = args[0].split('.')[-1]
            if ext == 'svg':
                return MockResponse(200)
            return MockResponse(404)

        self.scraper._session.get = mock_retry_request
        self.scraper._write_image = mock.Mock(return_value=None)

        for path, url, expected in test_cases:
            actual = self.scraper._retry_download_image(path, url)
            self.assertEqual(actual, expected)

    def test_retry_download_image_exception(self):
        test_cases = (
                      (# Test exhaustive retries with an unsupported img format
                       '001.test',
                       'http://gallery.com/g/abc/001.test',
                       ExtractorFailed
                      ),
                     )

        # Set up the mock functions
        def mock_retry_request(*args, **kwargs):
            # Mock the requests.get, only a 'png' img will return 200
            class MockResponse:
                def __init__(self, status_code):
                    self.status_code = status_code
            return MockResponse(404)

        self.scraper._session.get = mock_retry_request
        self.scraper._write_image = mock.Mock(return_value=None)

        for path, url, expected in test_cases:
            self.assertRaises(ExtractorFailed,
                              self.scraper._retry_download_image,
                              path,
                              url)

if __name__ == "__main__":
    unittest.main()



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains the abstract Extractor class"""

import re
import abc
import requests
from bs4 import BeautifulSoup


class Extractor(abc.ABC):
    """
    Abstract class outlining the core scraping functions for any website module
    to implement. All abstract public functions must be implemented as they
    are called by scraper.py

    It is recommended to overwrite these variables in the child classes:
        self.__PAGE_PATTERN      - General URL pattern for the site, used in
                                   is_page()
        self.__GALLERY_PATTERN   - URL pattern for a site gallery, used in
                                   is_gallery()

    Children may sometimes need to overwrite is_page() and is_gallery().
    """
    def __init__(self):
        self._PAGE_PATTERN = r""
        self._GALLERY_PATTERN = r""
        self._session = requests.Session()
        self._session.headers = {"User-Agent": "erodl/0.1",
                                 "Accept-encoding": "gzip"}
        requests.packages.urllib3.disable_warnings()

    def _get_soup(self, url):
        """Returns a BS4 soup from the URL's response.

        This is meant for internal use by classes that implement Extractor.
        """
        request = self._session.get(url, verify=False)
        request.raise_for_status()
        content = request.content
        return BeautifulSoup(content, "html.parser")

    # =========================================================================
    # Testing
    # =========================================================================

    def get_tests(self):
        """Return a tuple of dictionaries representing tests.

        The dictionaries are of the following format:
            {
             "url": "http://extractor-site.com/gallery/blahblah",
             "img_urls": [
                          "http://data.extractor.com/img/blahblah/001.jpg",
                          "http://data.extractor.com/img/blahblah/002.jpg",
                          ...,
                         ],
             "size": 24,
             "tags": {
                      "Title": "Tales of Blah Blah",
                      "Languages": "English",
                      "Artists": "Steve from Minecraft",
                      ...,
                     },
            }

        For 'img_urls', it is not necessary to list all URLS or the filenames.
        However, the value should at least contain the first few image URLs in
        incrementing order.

        Similarly for the 'tags', not all key-value pairs need to exist. The
        values can either be strings or lists of strings since the test suite
        flattens all values for comparison.

        See the abstract methods below for further details on proper formats
        for the values.

        Implementation of this method in subclasses is optional.
        """

    # =========================================================================
    # State checks
    # =========================================================================

    def is_page(self, url):
        """Return true if the extractor corresponds to the URL's site.

        This is a more generic check compared to is_gallery() as it only checks
        if the URL is for the site the extractor supports (aka. checks the
        hostname).
        """
        return bool(re.match(self._PAGE_PATTERN, url))

    def is_gallery(self, url):
        """Return true if the extractor can process the URL.

        This is a more precise check compared to is_page(), as it must check
        the URI pattern to see if the URL is a gallery (and thus can be
        scraped).
        """
        return bool(re.match(self._GALLERY_PATTERN, url))

    # =========================================================================
    # State modifiers
    # =========================================================================

    def reset(self):
        """Reset the extractor's state.

        Resets may be necessary for subclass implementations that rely on
        private variables to store results of complex calculations or lengthy
        requests. By clearing those variables, it prepares the extractor
        instance for scraping the next URL.

        Implementation of this method in subclasses is optional.
        """

    # =========================================================================
    # Getters
    # =========================================================================

    @abc.abstractmethod
    def get_size(self, url, soup, args):
        """Return number of images in gallery."""

    @abc.abstractmethod
    def get_gallery_urls(self, url, soup, args):
        """Return a list of image URLs and the image path to download to."""

    @abc.abstractmethod
    def get_tags(self, url, soup, args):
        """Return a dictionary of tags.

        Tags contain information regarding the gallery, such as the title,
        URL, language, artists, etc.

        Keys can be any string, but should at a minimum the dictionary should
        include: ("Title", "Languages", "Artists"/"Authors"/"Groups")

        Values are expected to be either strings, or lists containing strings.

        The following is an example tag dictionary:
        {
            "Title"        : "The Frog Prince",
            "Languages"    : "English",
            "Artists"      : ["Bob Ross", "Picasso"],
            "Groups"       : ["FSF", "Linux Foundation"],
            "Content"      : ["Amphibians", "Fantasy", "Adventure"],
            "URL"          : "http://www.books.com/gallery/the-frog-prince",
        }
        """



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module loads all the extractor classes for use by KomiDL"""

from .ehentai import EHentaiEX
from .hentai2read import Hentai2ReadEX
from .imgur import ImgurEX
from .mangadex import MangaDexEX
from .mangahub import MangaHubEX
from .mangasushi import MangaSushiEX
from .nhentai import NHentaiEX
from .tsumino import TsuminoEX



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module is the starting point for KomiDL where arguments are parsed and
passed to an instance of a KomiDL object
"""

import os
import sys
import argparse

import komidl.constants as constants
from komidl.komidl import KomiDL


def _init_args():
    """Parse the arguments given"""
    parser = argparse.ArgumentParser(prog="komidl",
                                     usage="%(prog)s [OPTIONS] URL [URL...]",
                                     description="A downloader for image \
                                                  galleries",
                                     epilog="Author: DNSheng",
                                     add_help=False,
                                     formatter_class=lambda prog:
                                     argparse.HelpFormatter(
                                         prog,
                                         max_help_position=90)
                                     )

    # =========================================================================
    # General options
    # =========================================================================
    general_group = parser.add_argument_group('General Options')
    general_group.add_argument("-h", "--help",
                               action="help",
                               help="Print this help menu")
    general_group.add_argument("-v", "--version",
                               action="version",
                               version=constants.VERSION,
                               help="Print program version")
    general_group.add_argument("-y", "--yes",
                               action="store_true",
                               help="Yes to all prompts")
    general_group.add_argument("--list-extractors",
                               action="store_true",
                               help="List all supported extractors")
    general_group.add_argument("--logo",
                               action="store_true",
                               help="Print the KomiDL logo")

    # =========================================================================
    # Download options
    # =========================================================================
    download_group = parser.add_argument_group('Download Options')
    download_group.add_argument("--tags",
                                action="store_true",
                                help="Download gallery tags to a text file")
    download_group.add_argument("--directory",
                                type=str,
                                default=".",
                                help="Download to a specific directory")
    download_group.add_argument("--lang",
                                type=str,
                                default="EN",
                                help="Choose the preferred language with \
                                      ISO 639-1 codes if option for site \
                                      exists - case-insensitive (default: EN)")
    download_group.add_argument("--thread-size",
                                type=int,
                                default=3,
                                help="Number of threads to download a gallery \
                                      with (default: 3)")

    # =========================================================================
    # Export options
    # =========================================================================
    export_group = parser.add_argument_group('Export Options')
    export_group.add_argument("--archive",
                              type=str,
                              choices=constants.ARCHIVE_FORMATS,
                              help="Archive the downloaded gallery")
    export_group.add_argument("--pdf",
                              action="store_true",
                              help="Export images downloaded to a PDF")
    export_group.add_argument("--keep",
                              action="store_true",
                              help="Keep gallery after exporting")

    # =========================================================================
    # URL input
    # =========================================================================
    parser.add_argument("URL",
                        type=str,
                        nargs="*",
                        help="One or more gallery/page URLs")

    args = parser.parse_args()

    # Custom parsing done for mutual exclusivity
    if not args.list_extractors and not args.URL and not args.logo:
        print("usage: komidl [OPTIONS] URL [URL...]")
        print("")
        print("komidl: error: You must provide at least one URL.")
        print("Type komidl --help to see a list of all options")
        sys.exit(0)

    return _preprocess_args(args)


def _preprocess_args(args):
    """Pre-process args before sending them to KomiDL"""
    # Return a tuple of similar language codes/abbreviations
    if args.lang.lower() in constants.ISO_LANG_SET:
        args.lang = constants.ISO_LANG_SET[args.lang.lower()]
    else:
        args.lang = (args.lang.lower(),)

    # Set the destination folder to an absolute path
    args.directory = os.path.abspath(args.directory)

    return args


def main():
    """Init args and pass to KomiDL for downloading"""
    args = _init_args()
    komidl = KomiDL(args)

    if args.list_extractors:
        komidl.list()
    elif args.logo:
        print(constants.KOMIDL_LOGO)
    else:
        try:
            komidl.download(args.URL)
        except KeyboardInterrupt:
            print("\nKeyboard interrupt detected -- exiting komidl")
            sys.exit(0)


if __name__ == '__main__':
    main()



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains the MangaHub extractor class"""

import os
import re
import ast
import json

from .extractor import Extractor


class MangaHubEX(Extractor):
    """
    An extractor for MangaHub.io

    In MangaHub, series URLs and chapter URLs are supported.
        + Chapter URLs are for a single chapter of a comic.
        + Series URLs include all chapters of a comic.

    The image URLs are retrieved by making a POST request to MangaHub's API and
    parsing the JSON. The request is built from a template and information
    from the URL.

    Tags come from parsing the comic's series page.

    Getting the chapter URLs and size requires extra requests to the server
    and thus are cached once retrieved.
    """
    def __init__(self):
        super().__init__()
        self.name = "MangaHub"
        self.url = "https://www.mangahub.io"
        self._PAGE_PATTERN = r"https?://(?:www\.)?mangahub\.io.*"
        self._GALLERY_PATTERN = r"https?://(?:www\.)?mangahub\.io/chapter/" + \
                                r"[a-z0-9-]+/?"
        self._SERIES_PATTERN = r"https?://(?:www\.)?mangahub\.io/manga/" + \
                               r"[a-z-]+/?"
        self._IMG_DOMAIN = "https://cdn.mangahub.io/file/imghub/"
        self._POST_API = "https://api2.mangahub.io/graphql"
        self._chapter_urls = None
        self._size = None

    # =========================================================================
    # State checks
    # =========================================================================

    def is_gallery(self, url):
        return self._is_chapter(url) or self._is_series(url)

    def _is_chapter(self, url):
        return re.match(self._GALLERY_PATTERN, url)

    def _is_series(self, url):
        return re.match(self._SERIES_PATTERN, url)

    # =========================================================================
    # State modifiers
    # =========================================================================

    def reset(self):
        self._chapter_urls = None
        self._size = None

    # =========================================================================
    # Getters
    # =========================================================================

    @staticmethod
    def _get_title(soup):
        """Get title of manga from main page"""
        rough = soup.title.string.split(" ")[1:-4]
        return " ".join(rough)

    @staticmethod
    def _get_categories(soup):
        category_tags = soup.find_all("a", {"class": "label genre-label"})
        return [item.string for item in category_tags]

    @staticmethod
    def _get_status(soup):
        return soup.find("span", {"class": "_3SlhO"},
                         string="Status").next_sibling.string

    @staticmethod
    def _get_artists(soup):
        return soup.find("span", {"class": "_3SlhO"},
                         string="Artist").next_sibling.string

    @staticmethod
    def _get_authors(soup):
        return soup.find("span", {"class": "_3SlhO"},
                         string="Author").next_sibling.string

    @staticmethod
    def _get_chapter_urls(soup):
        """Returns URLs for each chapter of a series"""
        latest_tag = soup.find("span", {"class": "_3SlhO"},
                               string="Latest").next_sibling
        latest_url = latest_tag.a.get('href')
        latest_num = latest_url.split("-")[-1]
        base_url = "-".join(latest_url.split("-")[:-1])
        return (f"{base_url}-{num}" for num in range(1, int(latest_num)+1))

    @staticmethod
    def _gallery_slug(gallery_url):
        """Return the gallery's slug and number for POST requests"""
        if gallery_url[-1] == "/":
            gallery_url = gallery_url[:-1]
        series = gallery_url.split('/')[-2]
        id_ = re.search("[0-9]+/?$", gallery_url)
        return series, id_.group(0)

    def _get_chapters(self, url, soup):
        """Return a string to summarize chapters scraped from URL"""
        if self._is_series(url):
            tag = soup.find("span", {"class": "_3SlhO"},
                            string="Latest").next_sibling
            number = tag.a.get("href").split("-")[-1]
            return f"0001-{number.zfill(4)}"
        number = url.split("-")[-1]
        return number.zfill(4)

    def _pages_dict(self, url):
        """Return a dictionary of pages for a chapter URL

        Data specific to the chapter is retrieved and used for a POST operation
        to the site's API. The JSON response is then parsed and returned.

        Returns a dict mapping image # to image URL
        """
        # Extract data from URL for POST
        slug, number = self._gallery_slug(url)
        query = f'{{chapter(x:m01,slug:"{slug}",number:{number}){{pages}}}}'
        data = {"query": query}
        # POST operation
        response = self._session.post(self._POST_API, data=data)
        # Convert the JSON to a python dict and grab pages
        json_data = json.loads(response.content.decode("utf-8"))
        literal_dict = json_data["data"]["chapter"]["pages"]
        return ast.literal_eval(literal_dict)

    def get_size(self, url, soup, args):
        """Get total # of images in the gallery to download

        The size is also saved in an attribute for quick retrieval later.
        """
        if self._size is None:
            if self._is_series(url):
                chapter_urls = (self._chapter_urls or
                                self._get_chapter_urls(soup))
                self._size = self._series_size(chapter_urls)
            else:
                self._size = self._chapter_size(url)
        return self._size

    def _series_size(self, chapter_urls):
        """Get total # of images in the series (the sum of all chapters)"""
        return sum(self._chapter_size(chapter) for chapter in chapter_urls)

    def _chapter_size(self, url):
        """Get total # of images of a chapter"""
        pages_dict = self._pages_dict(url)
        size_str = list(pages_dict.keys())[-1]
        return int(size_str)

    def get_gallery_urls(self, url, soup, args):
        """Return all image paths and image URLs"""
        if self._is_series(url):
            return self._get_series_images(soup)
        return self._get_chapter_images(url, soup)

    def _get_series_images(self, soup):
        """Return all images and image URLs for a series URL

        Combines the URLs for all chapters in the series.
        """
        series_urls = []
        chapter_urls = self._chapter_urls or self._get_chapter_urls(soup)
        for chapter in chapter_urls:
            chapter_num = chapter.split("-")[-1]
            ch_path = "Chapter " + chapter_num.zfill(4)
            series_urls += self._get_chapter_images(chapter, soup,
                                                    ch_path=ch_path)
        return series_urls

    def _get_chapter_images(self, url, soup, ch_path=None):
        """Return all images and image URLs for a chapter URL"""
        url_list = []
        size = self._size or self.get_size(url, soup, None)
        size_len = len(str(size))
        pages_dict = self._pages_dict(url)
        for img_num, sub_url in pages_dict.items():
            img_ext = sub_url.split(".")[-1]
            img_name = f"{img_num.zfill(size_len)}.{img_ext}"
            filename = (img_name if ch_path is None else
                        os.path.join(ch_path, img_name))
            img_url = f"{self._IMG_DOMAIN}{sub_url}"
            url_list.append([filename, img_url])
        return url_list

    def _main_soup(self, url):
        """Returns the soup of the manga series page"""
        if url[-1] == "/":
            url = url[:-1]
        sub_url, _ = os.path.split(url)
        main_url = re.sub(r"chapter", "manga", sub_url)
        return self._get_soup(main_url)

    def get_tags(self, url, soup, args):
        """Returns a dictionary of data regarding the series

        A majority of the series data is taken from the series page.
        Thus if a chapter URL is given, the series page needs to be scraped
        too (the 'main_soup').
        """
        main_soup = soup if self._is_series(url) else self._main_soup(url)

        tags = {"Languages": "English", "URL": url}
        tags["Title"] = self._get_title(main_soup)
        tags["Category"] = self._get_categories(main_soup)
        tags["Status"] = self._get_status(main_soup)
        tags["Artists"] = self._get_artists(main_soup)
        tags["Authors"] = self._get_authors(main_soup)
        tags["Chapters"] = self._get_chapters(url, soup)

        return tags



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains the EHentai extractor class"""

import re
from multiprocessing.dummy import Pool as ThreadPool

from .extractor import Extractor


class EHentaiEX(Extractor):
    """
    An extractor for E-Hentai.org
    """
    def __init__(self):
        super().__init__()
        self.name = "E-Hentai"
        self.url = "https://www.e-hentai.org"
        self._PAGE_PATTERN = r"https?://(?:www\.)?e-hentai\.org.*"
        self._GALLERY_PATTERN = r"https?://(?:www\.)?e-hentai\.org/g/" + \
                                r"[0-9]+/[a-z0-9]+/?"

    # =========================================================================
    # Getters
    # =========================================================================

    @staticmethod
    def _get_title(soup):
        title = soup.title.string
        if "]" in title:
            title = title.split("]")[1]
        if "(" in title:
            if title[0] == "(":
                title = title.split("(")[1].split(")")[1]
            else:
                title = title.split("(")[0]

        delims = ("|", "[")
        for delim in delims:
            if delim in title:
                title = title.split(delim)[0]
        return title.strip()

    @staticmethod
    def _get_id(url):
        return url.split("/")[-3] if url[-1] == "/" else url.split("/")[-2]

    @staticmethod
    def _get_language(soup):
        tags = soup.find_all("td", {"class": "gdt1"})
        lang_tag = [tag for tag in tags if tag.string == "Language:"][0]
        return lang_tag.next_sibling.next_element.strip()

    @staticmethod
    def _get_display_block(soup):
        gpc = soup.find("p", {"class": "gpc"}).string.split("-")[1]
        return gpc.strip().split(" ")[0]

    def _get_sub_urls(self, soup, url):
        """E-Hentai has an intermediate URL to go through before loading the
           image

        Parameters
        ----------
        soup : BS4Soup
            The soup of the original URL gallery
        url : str
            The URL of the gallery

        Returns
        -------
        list
            A list of all intermediate URLs
        """
        sub_urls = []
        block = self._get_display_block(soup)
        size = self.get_size(url, soup, None)
        # Ceiling divide size/block
        pages = -(-size // int(block))
        for page in range(0, pages):
            page_url = f"{url}?p={page}"
            soup = self._get_soup(page_url)
            a_tags = soup.find_all("a")
            gallery_id = self._get_id(url)
            sub_tags = [link.get("href") for link in a_tags
                        if link.get("href") and gallery_id in link.get("href")]
            sub_urls += [url for url in sub_tags
                         if re.match(f".*{gallery_id}-[0-9]+$", url)]
        return sub_urls

    def _retrieve_img_url(self, url):
        """From the sub URL, get the image URL

        This is called by the pool starmap

        Returns
        -------
        str
            The image URL
        """
        soup = self._get_soup(url)
        return soup.find("img", {"id": "img"})["src"]

    def _get_img_urls(self, soup, url, args):
        sub_urls = self._get_sub_urls(soup, url)
        pool = ThreadPool(args.thread_size)
        img_urls = pool.map(self._retrieve_img_url, sub_urls)
        return img_urls

    def get_size(self, url, soup, args):
        tags = soup.find_all("td", {"class": "gdt2"})
        tags = [item.string for item in tags]
        pages = [item for item in tags
                 if item is not None and "pages" in item][0]
        return int(pages.split(" ")[0])

    def get_gallery_urls(self, url, soup, args):
        url_list = []
        size = self.get_size(url, soup, args)
        size_len = len(str(abs(size)))
        img_urls = self._get_img_urls(soup, url, args)
        for img_num in range(1, size+1):
            img_type = img_urls[img_num-1].split(".")[-1]
            base_name = str(img_num).zfill(size_len)
            filename = f"{base_name}.{img_type}"
            url_list.append([filename, img_urls[img_num-1]])
        return url_list

    def get_tags(self, url, soup, args):
        soup_tags = {"Title": self._get_title(soup),
                     "Languages": self._get_language(soup),
                     "Artists": [],
                     "Groups": [],
                     "URL": url}
        key_tags = soup.find_all("td", {"class": "tc"})
        for key in key_tags:
            key_name = key.string.title()[:-1]
            if key_name == "Languages":
                continue
            if key_name == "Artist":
                key_name += "s"
            items = key.next_element.next_element.find_all("a")
            tag_list = self._clean_tags(items)
            soup_tags[key_name] = tag_list
        return soup_tags

    # =========================================================================
    # Misc. functions
    # =========================================================================

    @staticmethod
    def _clean_tags(items):
        """Capitalize every word"""
        return [item.string.title() for item in items]



from setuptools import setup, find_packages

import komidl.constants as constants

setup(
    # Application name:
    name="KomiDL",
    # Version number
    version=constants.VERSION,
    description="A downloader for image galleries and pages",
    # Tests
    test_suite="tests",
    # Packages
    packages=find_packages(),

    # Application author details:
    author="Dan Sheng",
    author_email="dnsheng@gmail.com",
    url="https://gitlab.com/dnsheng/komidl",

    # Include additional files into the package
    include_package_data=True,

    license="LICENSE",
    long_description=open("README.md").read(),

    # Dependencies
    python_requires='>3.6.0',

    # Dependent packages (distributions)
    install_requires=[
        "requests",
        "beautifulsoup4",
        "Pillow",
    ],

    entry_points={
        'console_scripts': [
            'komidl=komidl.__main__:main',
        ],
    },

    classifiers=[
        "Environment :: Console",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: MacOS",
        "Operating System :: Microsoft",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3.6",
        "Topic :: Multimedia :: Graphics",
    ],
)



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module will generate a list of all extractor instances for KomiDL"""

from .extractors import *

_ALL_CLASSES = [
        klass
        for name, klass in globals().items()
        if name.endswith("EX")
        ]


def _gen_extractor_classes():
    return _ALL_CLASSES


def get_extractors():
    """Return a list of all extractors"""
    return [klass() for klass in _gen_extractor_classes()]



"""Tests the extractor modules in KomiDL"""

import os
import sys
import unittest
from unittest.mock import patch
sys.path.append(os.path.abspath('..'))

from komidl.extractors import get_extractors

class ExtractorTest(unittest.TestCase):
    """Test the extractor modules in KomiDL"""

    def setUp(self):
        self.extractors = get_extractors()

    def _compare_size(self, test, extractor, soup):
        """Compare the expected size from the test dictionary to the actual
        size returned from the extractor implementation"""
        expected_size = test.get('size', None)
        if expected_size is not None:
            actual_size = extractor.get_size(test['url'], soup, None)
            self.assertEquals(expected_size, actual_size)

    def _compare_tags(self, test, extractor, soup):
        """Compare the expected tags from the test dictionary to the actual
        tags returned from the extractor implementation.

        If there exist key-value pairs in the actual implementation not found
        in the expected dictionary for the test, the test will still pass.
        """
        expected_tags = test.get('tags', None)
        if expected_tags is not None:
            actual_tags = extractor.get_tags(test['url'], soup, None)
            # Compare values, flatten potential lists to strings
            for key, value in expected_tags.items():
                expected_str = ''.join(value)
                actual_str = ''.join(actual_tags[key])
                # Normalize the two strings for easy comparison
                expected = sorted(expected_str.upper())
                actual = sorted(actual_str.upper())
                self.assertEquals(expected, actual)

    def _compare_urls(self, test, extractor, soup):
        """Compare the expected URLs from the test dictionary to the actual
        URLs returned from the extractor implementation.

        If there exist image URLs in the actual implementation not found in the
        expected values for the test, the test will still pass.
        """
        expected_imgs = test.get('img_urls', None)
        if expected_imgs is not None:
            actual_imgs = extractor.get_gallery_urls(test['url'], soup, None)
            # It's not expected of the tests to include all URLs and filenames
            # Thus, zip with the expected_imgs as the first arg to iterate
            for expected, (_, actual) in zip(expected_imgs, actual_imgs):
                self.assertEquals(expected, actual)

    def test_extractors(self):
        self.maxDiff = None
        for extractor in self.extractors:
            tests = extractor.get_tests()

            # Skip if no tests
            if not tests:
                continue

            for test in tests:
                url = test.get("url", None)
                # Check that there exists a URL key in the test
                self.assertTrue(url is not None)

                soup = extractor._get_soup(url)

                self._compare_size(test, extractor, soup)
                self._compare_tags(test, extractor, soup)
                self._compare_urls(test, extractor, soup)

if __name__ == "__main__":
    unittest.main()



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains the Imgur extractor class"""

from .extractor import Extractor


class ImgurEX(Extractor):
    """
    An extractor for Imgur.com

    When downloading galleries, Imgur supports various img/file formats.
    For a file '<ID>.gifv', '<ID>.jpg' also exists as a preview screenshot.
    Thus, it is unreliable to set all URLs as '.jpg' and rely on receiving an
    HTTP 404 error and letting Scraper's _retry_download_image() get the actual
    file. The workaround used is to use '<ID>.gifv' to request the HTML page
    for the iamge, and extract the true URL from there. The drawback is that
    the number of requests made for an album is doubled.

    Tags are non-existant for Imgur so the assumption is that the language is
    English.
    """
    def __init__(self):
        super().__init__()
        self.name = "Imgur"
        self.url = "https://www.imgur.com"
        self._PAGE_PATTERN = r"https?://(?:www\.)?imgur\.com.*"
        self._GALLERY_PATTERN = r"https?://(?:www\.)?imgur\.com/" + \
                                r"(?:a|gallery)/\w{5,7}/?"
        self._IMG_DOMAIN = "https://i.imgur.com/"

    # =========================================================================
    # Getters
    # =========================================================================

    @staticmethod
    def _get_title(url, soup):
        raw_title = soup.title.string.strip()
        title = '-'.join(raw_title.split('-')[:-1]).strip()
        if title == "Imgur: The magic of the Internet" or not title:
            # No title set by the uploader, use the Imgur gallery ID
            album_id = (url.split("/")[-1] if url[-1] != "/"
                        else url.split("/")[-2])
            title = f"Imgur-{album_id}"
        return title

    def get_size(self, url, soup, args):
        image_tags = soup.findAll("div", {"class": "post-image-container"})
        return len(image_tags)

    def _image_url(self, img_tag):
        """For an image tag, return the image's true URL with extension

        Using '.gifv', we can obtain an HTML page that hosts the
        image/video. Videos use the <source> tag and images use the
        <img> tag. Note that there will always be at least one <img> tag
        due to the favicon.
        """
        img_id = img_tag.get('id')
        site_url = f"{self._IMG_DOMAIN}{img_id}.gifv"

        image_soup = self._get_soup(site_url)
        img_url = image_soup.find("img")["src"]
        video_tag = image_soup.find("source")
        if video_tag:
            img_url = video_tag["src"]

        return f"https:{img_url}"

    def get_gallery_urls(self, url, soup, args):
        url_list = []
        size = self.get_size(url, soup, args)
        size_len = len(str(size))
        img_tags = soup.findAll("div", {"class": "post-image-container"})
        img_urls = (self._image_url(tag) for tag in img_tags)

        for img_num, img_url in enumerate(img_urls):
            base_name = str(img_num).zfill(size_len)
            extension = img_url.split('.')[-1]
            filename = f"{base_name}.{extension}"
            url_list.append([filename, img_url])

        return url_list

    def get_tags(self, url, soup, args):
        soup_tags = {"Title": self._get_title(url, soup),
                     "Languages": "English",
                     "URL": url}
        return soup_tags



# KomiDL - A gallery downloader
# Copyright (C) 2019 DNSheng
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module contains the KomiDL class"""

import os
import re
import sys
import shutil

from PIL import Image
from requests.exceptions import HTTPError

import komidl.constants as constants
from komidl.scraper import Scraper
from komidl.extractors import get_extractors
from komidl.exceptions import ExtractorFailed, InvalidURL


class KomiDL:
    """Match URLs to extractors and start the download procedure.

    This class is responsible for handling the URLs, and any actions
    before/after the scraping and downloading (ex. listing extractors, archive
    downloads, etc.).
    Furthermore, this class owns instances of all extractors and does matching
    between URL to extractor.

    """
    def __init__(self, args):
        self._args = args
        self._extractors = get_extractors()
        self._scraper = Scraper()

    @staticmethod
    def _check_protocol(url: str) -> str:
        """Return a URL with a protocol appended.

        If a protocol is missing from the URL, print a warning and return
        the URL using the HTTP protocol.
        """
        if not re.match(r"(?:http|https)://", url):
            print("Warning: No protocol specified in URL, trying HTTP")
            url = f"http://{url}"
        return url

    def _get_extractor(self, url: str):
        """Return the appropriate extractor for the URL.

        If no appropriate extractor is found, an ExtractorFailed exception
        is raised.
        """
        for extractor in self._extractors:
            if extractor.is_page(url):
                return extractor
        raise ExtractorFailed(f"Extractor not found for site: {url}")

    @staticmethod
    def _archive_dir(path: str, frm: str, dest: str) -> None:
        """Archive the directory and its contents to the destination folder.

        Parameters
        ----------
        path:   The path of the directory to archive
        frm:    The archive format to use
        dest:   The path to save the archive to
        """
        _, base = os.path.split(path)
        archive = f"{path}.{frm}"
        archive_pattern = rf"{re.escape(archive)}.*\.{frm}"
        old_path = path

        if os.path.isfile(archive):
            duplicates = sum(1 for file in os.listdir(dest)
                             if re.match(archive_pattern, file))
            old_path = path
            base = f"{base} ({duplicates})"
            path = f"{path} ({duplicates})"
            os.rename(old_path, path)

        shutil.make_archive(path, frm, root_dir=dest, base_dir=base)
        os.rename(path, old_path)

    @staticmethod
    def _export_pdf(path: str, dest: str) -> None:
        """Export all images from the directory to PDF in the destination dir.

        Parameters
        ----------
        path:   The path of the directory containing images
        dest:   The path to save the PDF to
        """
        _, folder_name = os.path.split(path)
        # Recursively traverse directory, get all file paths
        file_paths = (os.path.join(root, file)
                      for root, _, files in os.walk(path)
                      for file in files)

        # Get all valid image files from path and open with PIL
        valid_imgs = (file for file in file_paths
                      if file.split('.')[-1] in constants.IMAGE_FORMATS)
        pil_imgs = (Image.open(img) for img in valid_imgs)

        def convert_img(img):
            """Convert all PIL Image objects to RGB."""
            if img.mode != "RGB":
                # Filename is necessary for sequential image ordering
                filename = img.filename
                img = img.convert('RGB')
                img.filename = filename
            return img

        pil_rgb = (convert_img(img) for img in pil_imgs)

        # Order images by sorting with filename and separate out title image
        title_img, *content = sorted(pil_rgb, key=lambda img: img.filename)

        # Construct the PDF filename based on existing files in dest dir
        pdf_pattern = rf"{re.escape(folder_name)}.*\.pdf"
        duplicates = sum(1 for file in os.listdir(dest)
                         if re.match(pdf_pattern, file))

        pdf_filename = f"{folder_name}.pdf"
        if duplicates > 0:
            pdf_filename = f"{folder_name} ({duplicates}).pdf"

        # Export all images to PDF
        title_img.save(pdf_filename, "PDF", resolution=100.0, save_all=True,
                       append_images=content)

    def list(self) -> None:
        """List all extractors"""
        for e in self._extractors:
            name_str = f"[{e.name}]"
            print(f"{name_str:<18}| {e.url}")

    def download(self, urls) -> None:
        """Download all images from the list of URLs.

        For each URL, an appropriate extractor is found and used to download
        all images and tags. After downloading, export operations may be done
        if specified in the runtime options (args).

        Parameters
        ----------
        urls : list [str]
            A list of gallery URLs to download
        """
        for url in urls:
            url = self._check_protocol(url)
            try:
                self._scraper.extractor = self._get_extractor(url)
                gallery = self._scraper.scrape(url, self._args)
            except (ValueError, InvalidURL, ExtractorFailed, HTTPError) as e:
                print(f"{e.__class__.__name__}: {e}")
                continue
            except (FileNotFoundError) as e:
                print(f"{e.__class__.__name__}: {e}")
                sys.exit(1)
            finally:
                self._scraper.reset()

            # Optional gallery export operations
            if self._args.pdf:
                self._export_pdf(gallery, self._args.directory)

            if self._args.archive:
                self._archive_dir(gallery, self._args.archive,
                                  self._args.directory)

            # If an export option was chosen, check if gallery should be kept
            if self._args.pdf or self._args.archive:
                if not self._args.keep:
                    shutil.rmtree(gallery)



"""Tests the status module in KomiDL"""

import os
import sys
import unittest
from unittest.mock import patch
sys.path.append(os.path.abspath('..'))

import komidl.status as status

class StatusTest(unittest.TestCase):
    """Test the status module in KomiDL"""
    #######################################################################
    # yn_prompt()
    #######################################################################

    def test_yn_lower_y(self):
        """Lowercase 'y' entered in yn_prompt()"""
        with patch("builtins.input", return_value='y') as _input:
            self.assertEqual(status.yn_prompt("Test msg"), True)
    
    def test_yn_upper_y(self):
        """Uppercase 'Y' entered in yn_prompt()"""
        with patch("builtins.input", return_value='Y') as _input:
            self.assertEqual(status.yn_prompt("Test msg"), True)
    
    def test_yn_lower_n(self):
        """Lowercase 'n' entered in yn_prompt()"""
        with patch("builtins.input", return_value='n') as _input:
            self.assertEqual(status.yn_prompt("Test msg"), False)

    def test_yn_upper_n(self):
        """Uppercase 'N' entered in yn_prompt()"""
        with patch("builtins.input", return_value='N') as _input:
            self.assertEqual(status.yn_prompt("Test msg"), False)
    
    def test_yn_random(self):
        """A character not 'y' or 'n' entered in yn_prompt()"""
        with patch("builtins.input", return_value='q') as _input:
            self.assertEqual(status.yn_prompt("Test msg"), False)
    
    def test_yn_random(self):
        """No character entered in yn_prompt()"""
        with patch("builtins.input", return_value='') as _input:
            self.assertEqual(status.yn_prompt("Test msg"), False)

    def test_yn_yes_arg(self):
        """Skip the prompt if args.yes is True"""
        self.assertEqual(status.yn_prompt("Test msg", True), True)

    #######################################################################
    # Miscellaneous
    #######################################################################
    
    def test_get_bars(self):
        """Test # of bars to display"""
        test_cases = ((50, 100, 50, 25),
                      (50, 100, 33, 16),
                      (50, 100, 100, 50),
                      (50, 100, 0, 0))
        for barsize, gallery_size, current, expected_bars in test_cases:
            actual_bars = status._get_bars(barsize, current, gallery_size)
            self.assertEqual(expected_bars, actual_bars)

    def test_get_percent(self):
        """Test percentage calculations"""
        test_cases = ((100, 50, 50),
                      (100, 23, 23),
                      (100, 100, 100),
                      (100, 0, 0))
        for gallery_size, current, expected_percent in test_cases:
            actual_percent = status._get_percent(current, gallery_size)
            self.assertEqual(expected_percent, actual_percent)

if __name__ == "__main__":
    unittest.main()
